//
// Virtual Trackball based on Quaternions
//
// Reto Ltolf,1999,2001.
// E-mail: luetolf@inf.ethz.ch
// URL: http://graphics.ethz.ch/
//


#include "trackball.h"


//globals
static GLint     tb_timestamp;
static GLuint    tb_width;
static GLuint    tb_height;
static GLuint    tb_button  = GLUT_MIDDLE_BUTTON;
static GLboolean tb_animate = GL_FALSE;
static GLboolean tb_rotate  = GL_FALSE;
static GLfloat   tb_matrix[4][4];
static GLfloat   tb_lastquat[4];
static GLfloat   tb_newquat[4];
static GLfloat   tb_lastpos[3];


//internal vector functions
void _TB_VEC_copy(const GLfloat *v1, GLfloat *v2)
{
  register int i;

  for (i=0;i<3;i++) v2[i] = v1[i];
}

void _TB_VEC_scale(GLfloat *v, GLfloat s)
{
  v[0] *= s;
  v[1] *= s;
  v[2] *= s;
}

void _TB_VEC_add(const GLfloat *v1, const GLfloat *v2, GLfloat *s)
{
  s[0] = v1[0] + v2[0];
  s[1] = v1[1] + v2[1];
  s[2] = v1[2] + v2[2];
}

GLfloat _TB_VEC_dot(const GLfloat *v1, const GLfloat *v2)
{
  return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

void _TB_VEC_cross(const GLfloat *v1, const GLfloat *v2, GLfloat *cross)
{
  GLfloat t[3];

  t[0] = (v1[1]*v2[2]) - (v1[2]*v2[1]);
  t[1] = (v1[2]*v2[0]) - (v1[0]*v2[2]);
  t[2] = (v1[0]*v2[1]) - (v1[1]*v2[0]);
  _TB_VEC_copy(t, cross);
}

void _TB_VEC_project(GLfloat v[3], int x, int y)
{
  GLfloat x_coord;
  GLfloat y_coord;

  x_coord = (2.0*x)/(GLfloat)tb_width - 1.0;
  y_coord = 1.0 - (2.0*y)/(GLfloat)tb_height;

  //if (fabs(x_coord)>1) x_coord = copysign(1.0, x_coord);
  //if (fabs(y_coord)>1) y_coord = copysign(1.0, y_coord);

  if (x_coord>1.0)  x_coord =  1.0;
  if (x_coord<-1.0) x_coord = -1.0;
  if (y_coord>1.0)  y_coord =  1.0;
  if (y_coord<-1.0) y_coord = -1.0;

  v[0] = x_coord;
  v[1] = y_coord;
  v[2] = sqrt(TB_TRACKBALL_RADIUS*TB_TRACKBALL_RADIUS - x_coord*x_coord - y_coord*y_coord);
}


//internal quaternion functions
void _TB_QUAT_norm(GLfloat quat[4])
{
  register int i;
  GLfloat magnitude;

  magnitude = sqrt(quat[0]*quat[0] + quat[1]*quat[1] + quat[2]*quat[2] + quat[3]*quat[3]);
  for (i=0;i<4;i++) quat[i] /= magnitude;
}

void _TB_QUAT_mult(GLfloat quat1[4], GLfloat quat2[4], GLfloat dest[4])
{
  GLfloat t1[4], t2[4], t3[4], tf[4];

  _TB_VEC_copy(quat1,t1); _TB_VEC_scale(t1,quat2[3]);
  _TB_VEC_copy(quat2,t2); _TB_VEC_scale(t2,quat1[3]);

  _TB_VEC_cross(quat1,quat2,t3);	
  _TB_VEC_add(t1,t2,tf); _TB_VEC_add(t3,tf,tf);

  tf[3] = quat1[3]*quat2[3] - _TB_VEC_dot(quat1,quat2);

  dest[0] = tf[0];
  dest[1] = tf[1];
  dest[2] = tf[2];
  dest[3] = tf[3];
  _TB_QUAT_norm(dest);
}

void _TB_QUAT_buildQuat(GLfloat quat[4], GLfloat v[3])
{
  GLfloat a[3];
  GLfloat dist[3];
  GLfloat phi;

  if (v[0]==tb_lastpos[0] && v[1]==tb_lastpos[1]) {
    quat[0] = 0.0;
    quat[1] = 0.0;
    quat[2] = 0.0;
    quat[3] = 1.0;
    return;
  }

  dist[0] = v[0] - tb_lastpos[0];
  dist[1] = v[1] - tb_lastpos[1];
  dist[2] = v[2] - tb_lastpos[2];

  phi = sqrt(dist[0]*dist[0]+dist[1]*dist[1]+dist[2]*dist[2]) / (2.0*TB_TRACKBALL_RADIUS);
  if (phi>1.0) phi = 1.0;
  if (phi<-1.0) phi = -1.0;
  phi = 2.0 * asin(phi);

  _TB_VEC_cross(tb_lastpos, v, a);
  _TB_VEC_scale(a, 1.0/(a[0]*a[0]+a[1]*a[1]+a[2]*a[2]));
  _TB_VEC_copy(a, quat);
  _TB_VEC_scale(quat, sin(phi/2.0));
  quat[3] = cos(phi/2.0);
}


void _TB_QUAT_buildMatrix(GLfloat quat[4])
{
  tb_matrix[0][0] = 1.0 - 2.0 * (quat[1]*quat[1] + quat[2]*quat[2]);
  tb_matrix[1][0] =       2.0 * (quat[0]*quat[1] - quat[2]*quat[3]);
  tb_matrix[2][0] =       2.0 * (quat[2]*quat[0] + quat[1]*quat[3]);
  tb_matrix[3][0] = 0.0;

  tb_matrix[0][1] =       2.0 * (quat[0]*quat[1] + quat[2]*quat[3]);
  tb_matrix[1][1] = 1.0 - 2.0 * (quat[2]*quat[2] + quat[0]*quat[0]);
  tb_matrix[2][1] =       2.0 * (quat[1]*quat[2] - quat[0]*quat[3]);
  tb_matrix[3][1] = 0.0;

  tb_matrix[0][2] =       2.0 * (quat[2]*quat[0] - quat[1]*quat[3]);
  tb_matrix[1][2] =       2.0 * (quat[1]*quat[2] + quat[0]*quat[3]);
  tb_matrix[2][2] = 1.0 - 2.0 * (quat[1]*quat[1] + quat[0]*quat[0]);
  tb_matrix[3][2] = 0.0;

  tb_matrix[0][3] = 0.0;
  tb_matrix[1][3] = 0.0;
  tb_matrix[2][3] = 0.0;
  tb_matrix[3][3] = 1.0;
}


//internal functions
static void _TB_idle_callback(void)
{
  _TB_QUAT_mult(tb_newquat, tb_lastquat, tb_lastquat);
  glutPostRedisplay();
}

void _TB_start(int x, int y, GLint time)
{
  tb_timestamp = time;
  tb_rotate = GL_TRUE;

  _TB_VEC_project(tb_lastpos, x, y);
}

void _TB_stop(GLint time)
{
  tb_rotate = GL_FALSE;
  if (time==tb_timestamp && tb_animate) glutIdleFunc(_TB_idle_callback);
  else if (tb_animate) glutIdleFunc(NULL);
}


//public functions
void TB_init(GLuint button)
{
  GLint current_mode;

  tb_button = button;

  tb_lastquat[0] = 0.0;
  tb_lastquat[1] = 0.0;
  tb_lastquat[2] = 0.0;
  tb_lastquat[3] = 1.0;    

  glGetIntegerv(GL_MATRIX_MODE, &current_mode);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glGetFloatv(GL_MODELVIEW_MATRIX, &tb_matrix[0][0]);
  glPopMatrix();
  glMatrixMode(current_mode);
}

void TB_animate(GLboolean animate)
{
  tb_animate = animate;
}

void TB_mouse(int button, int state, int x, int y)
{
  if (button==tb_button && state==GLUT_DOWN) _TB_start(x, y, glutGet(GLUT_ELAPSED_TIME));
  else if (button==tb_button && state==GLUT_UP) _TB_stop(glutGet(GLUT_ELAPSED_TIME));
}

void TB_motion(int x, int y)
{
  GLfloat new_pos[3];

  if (!tb_rotate) return;
  
  _TB_VEC_project(new_pos, x, y);
  _TB_QUAT_buildQuat(tb_newquat, new_pos);
  _TB_VEC_copy(new_pos, tb_lastpos);
  _TB_QUAT_mult(tb_newquat, tb_lastquat, tb_lastquat);

  tb_timestamp = glutGet(GLUT_ELAPSED_TIME);
  glutPostRedisplay();
}

void TB_reshape(int w, int h)
{
  tb_width = w;
  tb_height = h;
}

void TB_matrix(void)
{
  GLint current_mode;

  _TB_QUAT_buildMatrix(tb_lastquat);

  glGetIntegerv(GL_MATRIX_MODE, &current_mode);
  glMatrixMode(GL_MODELVIEW);
  glMultMatrixf(&tb_matrix[0][0]);
  glMatrixMode(current_mode);
}
